#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include "mandelbrot.h"
#include "colorizer.h"
#include "pixelbuf.h"

#define RGB(r, g, b) \
   {.red = (double)r / 255, .green = (double)g / 255, .blue = (double)b / 255}

RGBColor sunset[] = {
   RGB(25, 7, 26),
   RGB(9, 1, 47),
   RGB(4, 4, 73),
   RGB(0, 7, 100),
   RGB(12, 44, 138),
   RGB(24, 82, 177),
   RGB(57, 125, 209),
   RGB(134, 181, 229),
   RGB(211, 236, 248),
   RGB(241, 233, 191),
   RGB(248, 201, 95),
   RGB(255, 170, 0),
   RGB(204, 128, 0),
   RGB(153, 87, 0),
   RGB(106, 52, 3),
   RGB(66, 30, 15),
};

typedef struct {
   double center_x; double center_y;
   double horizontal_diameter;
   unsigned int width; unsigned int height; // in pixels
   unsigned int maxiterations;
   double color_shift;
   double color_progression;
   RGBColor* colors; unsigned int nof_colors;
   // derived parameters
   double resolution;
   double xmin;
   double xmax;
   double vertical_diameter;
   double ymin;
   double ymax;
} Parameters;

static Parameters p = {
   .center_x = -0.7,
   .center_y = 0,
   .horizontal_diameter = 3.07,
   .width = 256,
   .height = 192,
   .maxiterations = 1000,
   .color_shift = 0,
   .color_progression = 9,
   .colors = sunset,
   .nof_colors = sizeof(sunset)/sizeof(sunset[0]),
};

#define MEMBER(field) ((void*)&p.field)

bool read_double(char* arg, void* p) {
   return sscanf(arg, "%lf", (double*) p) == 1;
}

bool read_uint(char* arg, void* p) {
   return sscanf(arg, "%u", (unsigned int*) p) == 1;
}

static struct Options {
   char name;
   const char* description;
   bool (*input)(char* arg, void* p);
   void* addr;
} options[] = {
   {'x', "center (real part)", read_double, MEMBER(center_x)},
   {'y', "center (imaginary part)", read_double, MEMBER(center_y)},
   {'d', "horizontal diameter", read_double, MEMBER(horizontal_diameter)},
   {'w', "width in pixels", read_uint, MEMBER(width)},
   {'h', "height in pixels", read_uint, MEMBER(height)},
   {'m', "maximal # of iterations", read_uint, MEMBER(maxiterations)},
   {'s', "color shift", read_double, MEMBER(color_shift)},
   {'p', "color progression",
      read_double, MEMBER(color_progression)},
};
static const unsigned int nofoptions = sizeof(options)/sizeof(options[0]);

static char* cmdname = 0;

static void usage() {
   fprintf(stderr, "Usage: %s", cmdname);
   for (int i = 0; i < nofoptions; ++i) {
      fprintf(stderr, " [-%c value]", options[i].name);
   }
   fprintf(stderr, " outfile\n");
   for (int i = 0; i < nofoptions; ++i) {
      fprintf(stderr, "   -%c value: %s\n",
	 options[i].name, options[i].description);
   }
   exit(1);
}

static RGBColor get_color_of_mandelbrot(Colorizer* colorizer, Parameters* p,
      int row, int col) {
   double x = p->xmin + col/(double)p->resolution;
   double y = p->ymax - row/(double)p->resolution;
   return mandelbrot(x, y, p->maxiterations, colorizer);
}

int main(int argc, char** argv) {
   /* process arguments */
   cmdname = *argv++; --argc;
   while (argc > 0 && argv[0][0] == '-') {
      char option = argv[0][1];
      unsigned int i = 0;
      while (i < nofoptions && option != options[i].name) ++i;
      if (i == nofoptions) usage();
      char* val;
      if (argv[0][2]) {
	 val = &argv[0][2];
      } else {
	 --argc; ++argv;
	 if (argc == 0) usage();
	 val = *argv;
      }
      if (!options[i].input(val, options[i].addr)) usage();
      --argc; ++argv;
   }
   if (argc != 1) usage();
   char* outfile = *argv++; --argc;

   /* compute derived parameters */
   p.resolution = p.width / p.horizontal_diameter;
   p.xmin = p.center_x - p.horizontal_diameter / 2;
   p.xmax = p.center_x + p.horizontal_diameter / 2;
   p.vertical_diameter = p.height / p.resolution;
   p.ymin = p.center_y - p.vertical_diameter / 2;
   p.ymax = p.center_y + p.vertical_diameter / 2;

   /* set up colorizer */
   Colorizer colorizer = {
      .colors = p.colors,
      .nof_colors = p.nof_colors,
      .shift = p.color_shift,
      .progression = p.color_progression,
   };

   /* generate pixelbuf and dump it as JPEG file */
   PixelBuf pixelbuf;
   pixelbuf_init(&pixelbuf, p.width, p.height);
   for (unsigned int row = 0; row < p.height; ++row) {
      for (unsigned int col = 0; col < p.width; ++col) {
	 RGBColor color = get_color_of_mandelbrot(&colorizer, &p, row, col);
	 pixelbuf_set(&pixelbuf, row, col, color);
      }
   }
   pixelbuf_gen_jpeg(&pixelbuf, outfile);
   pixelbuf_free(&pixelbuf);
}
